#include <stdlib.h>
#include <super_malloc.h>
#include <stdio.h>
#include <string.h>
#include <list_node.h>
#include <debug.h>

/* Crea una nueva lista y la pone en *list
   Si free_node != NULL se llama cuando se libera un nodo con el nodo como arg
   Si free_data != NULL se llama cuando se libera un nodo con el dato como arg
*/
void list_head_new(list_head **list, int free_node, int (*free_data)(void *))
{
  *list = struct_malloc(list_head);
  list_head_init(*list, free_node, free_data);
}

void list_head_init(list_head *list, int free_node, int (*free_data)(void *))
{
  memset(list, 0, sizeof(list_head));
  list->free_node = free_node;
  list->free_data = free_data;
}

u_int32_t list_head_get_size(list_head *h)
{
  return h->size;
}
/* Libera una lista */
void list_head_free(list_head *list, int free_head)
{
  if (list->free_node) {
    list_head_iterator_next(list, list_node_free, list);
  }
  if (free_head) free(list);
}

/* Libera un nodo de la lista list_pointer */
int list_node_free(void *list_pointer, list_node *node)
{
  list_head *list = (list_head *)list_pointer;
  if (list->free_data) list->free_data(node->ld_vpointer);
  free(node);
  return 1;
}

/* Inserta un elemento al final de la lista. Complejidad O(1).
   Si node == NULL se crea un nuevo nodo con malloc (util cuando
   se especifica free_node y free_data) */
void list_head_push_integer(list_head *h, int d, list_node *node)
{
  list_head_push_any(h, d, NULL, node);
}

void list_head_push_vpointer(list_head *h, void *d, list_node *node)
{
  list_head_push_any(h, 0, d, node);
}

void list_head_push_any(list_head *h, int di, void *dv, list_node *node)
{
  if (node == NULL) node = struct_malloc(list_node);
  h->size++;
  if (dv == NULL) {
    node->ld_int = di;
  } else {
    node->ld_vpointer = dv; 
  }
  if (h->last == NULL) {
    h->last = h->first = node;
    node->next = node->prev = NULL;
    return;
  }
  node->prev = h->last;
  node->next = NULL;
  h->last->next = node;
  h->last = node;
}

/* Inserta |d| al inicio de la lista. Complejidad O(1)
   Si node == NULL se crea un nuevo nodo con malloc (util cuando
   se especifica free_node y free_data) */
void list_head_unshift_integer(list_head *h, int d, list_node *node)
{
  list_head_unshift_any(h, d, NULL, node);
}

void list_head_unshift_vpointer(list_head *h, void *d, list_node *node)
{
  list_head_unshift_any(h, 0, d, node);
}

void list_head_unshift_any(list_head *h, int di, void *dv, list_node *node)
{
  if (node == NULL) node = struct_malloc(list_node);
  h->size++;
  if (dv == NULL) {
    node->ld_int = di;
  } else {
    node->ld_vpointer = dv;
  }
  if (h->first == NULL) {
    h->last = h->first = node;
    node->next = node->prev = NULL;
    return;
  }
  node->prev = NULL;
  node->next = h->first;
  h->first->prev = node;
  h->first = node;
}

/* Borra un elemento al final de la lista*/
void* list_head_pop(list_head *h, int *ret)
{
  list_node *tmp, *out;
#ifdef _DEVELOP_VERSION
  if (h->last == NULL) {
    fprintf(stderr,"list_head_pop: no elements\n");
    return NULL;
  }
#endif
  h->size--;
  out = h->last;
  *ret = out->ld_int;
  if (h->last == h->first) {
    if (h->free_node) list_node_free(h, h->last);
    h->last = h->first = NULL;
    return out;
  }
  tmp = h->last->prev;
  if (tmp != NULL) tmp->next = NULL;
  if (h->free_node) list_node_free(h, h->last);
  h->last = tmp;
  return out;
}

/* Borra un elemento al principio de la lista*/
void * list_head_shift(list_head *h, int *ret)
{
  list_node *tmp, *out;
#ifdef _DEVELOP_VERSION
  if (h->first == NULL) {
    fprintf(stderr,"list_head_shift: no elements\n");
    return 0;
  }
#endif
  h->size--;
  out = h->first;
  *ret = out->ld_int;
  if (h->last == h->first) {
    if (h->free_node) list_node_free(h, h->first);
    h->last = h->first = NULL;
    return out;
  }
  tmp = h->first->next;
  if (tmp != NULL) tmp->prev = NULL;
  if (h->free_node) list_node_free(h, h->first);
  h->first = tmp;
  return out;
}

/* Borra el nodo |n| de la lista |h|*/
void list_head_remove_node(list_head *h, list_node *n)
{
  if (n->prev != NULL) {
    n->prev->next = n->next;
  } else {
    h->first = n->next;
  }
  if (n->next != NULL) {
    n->next->prev = n->prev;
  } else {
    h->last = n->prev;
  }
  h->size--;
}

/* Recorre hacia adelante sobre los elementos de una lista*/
void list_head_iterator_next(list_head *h,
			     int (*visit)(void *, list_node *), void *arg)
{
  register list_node *i, *j;
  i = h->first;
  while (i) {
    j = i;
    i = i->next;
    if (visit(arg, j) == 0) break;
  }
}

/* Recorre hacia atras sobre los elementos de una lista*/
void list_head_iterator_prev(list_head *h,
			int (*visit)(void *, list_node *), void *arg)
{
  list_node *i, *j;
  i = h->last;
  while (i) {
    j = i;
    i = i->prev;
    if (visit(arg, j) == 0) return;
  }
}

/* Imprime un nodo de la lista */
int list_node_print(void *arg, list_node *node)
{
  fprintf(stderr,"prev=%p, this=%p, data=%04d, next=%p\n",
	  node->prev,node,node->ld_int,node->next);
  return 1;
}
