#ifndef __H_QUERY__
#define __H_QUERY__

#include <stdlib.h>
#include <sys/types.h>
#include <iterator.h>
#include <idx_man.h>
#include <id_wrd_t.h>
#include <document_env.h>
#include <skip_list.h>

struct query {
  u_int8_t argc; /* Nmero de palabras */
  char **argv; /* Las palabras */
  idx_man *idm; /* Manejador de preguntas */
  iterator **__itrs; /* Iteradores de las palabras encontradas uso interno */
  iterator **itrs; /* Iteradores de las palabras encontradas */
  idx_global_wrd_data *gdata; /* Datos globales */
  document_env *env; /* */
  /* Nmero de documentos en todo el sistema. Si se cambia debe ser despus
     de las llamadas a query_init */
  u_int64_t collection_size;
};

typedef struct query query;

struct query_result {
  u_int16_t psize; /* Nmero de documentos posibles en el resultado */
  skip_list *docs; /* Puntero a documentos */
  /* Saca la informacin de |info| y pone el |id_doc_t| en el puntero
     id_doc, y pone la medida de similaridad en sim. */
  void (*get_info)(struct query_result *res, void *info,
		   id_doc_t **id_doc, float *sim);
  
  /* Qu hacer con elementos duplicados (Nosotros decimos que
     es duplicado en |cmp|. Regresa el elemento escogido. Si se pone a
     NULL no hay manejador especial para los duplicados y los nodos iguales
     se ponen uno tras de otro.
     Si los duplicados se manejan, se utiliza una lista |list_head|
     (ver list_node.h) para unirlos si dup_handler as lo requiere.
     - new_data es un list_head (ver list_node.h) y en los nodos de la lista
     el campo |data| contiene los punteros.
     - in_skip_list es un list_head. */
  void* (*dup_handler)(void *in_skip_list, void *new_data);
  /* Funcin de comparacin (que documento es ms parecido a la consulta)
     Si hay duplicados el elemento en la lista se accede como:
     (yourtype *)((list_head *)in_skip_list)->first->data
     Si no hay duplicados:
     (yourtype *)in_skip_list
     El nuevo elemento: 
     (yourtype *)new_data
     Si in_skip_list < new_data regresa negativo
     Si in_skip_list > new_data regresa numero positivo
     Si in_skip_list = new_data regresa 0 */
  int (*cmp)(void *in_skip_list, void *new_data);
  /* liberador de un elemento, debe regresar 1 a menos que se desee
     cancelar las liberaciones posteriores, solo cuando se llama desde free,
     al liberar la query_result (ver skip_list.h).
     Si se pone a NULL se utiliza free (man 3) regresando 1.*/
  int (*free_item)(void *data);
  /* libera la query_result */
  void (*free)(struct query_result *r);
  /* push data into result */
  void (*push)(struct query_result *r, void *data);
  /* remove the first element (bigger similarity on cmp) */
  void (*pop)(struct query_result *qr);
};

typedef struct query_result query_result;

struct query_result_iterator {
  query_result *qres;
  skip_list_iterator *itr;
};

typedef struct query_result_iterator query_result_iterator;

/* Crea una nueva consulta */
int query_new(query **qry, document_env *env,
	      idx_man *idm, char **argv, int argc);

/* Destruye una consulta */
void query_close(query *q);

/* reinicia los iteradores y pone los valores globales de nuevo */
void query_reset(query *q);

/* Inicializa y realiza una consulta, si init_all == 1 regresa
   1 si todas se encontraron, y 0 si no (regresa sin completar
   lo que falte). */
int query_init(query *q, int init_all);

/* crea un nuevo resultado */
void query_result_new(query_result **qr, int size,
		      void (*get_info)(query_result *, void *,
				       id_doc_t **, float *),
		      int (*cmp)(void *, void*),
		      int (*free_item)(void *));

/* libera un resultado */
void query_result_free(query_result *qr);

/* Inserta el dato 'data' en  'qr->docs' y si el dato existe lo remueve y lo
   vuelve a insertar, sigue utilizando 'dup' para hacer saber al dato
   anterior que se realizan cambios, se reinserta para reacomodarse a
   un posible cambio de peso del nodo. El tamao de la lista es regulado
   por 'max', para una lista no regulada 'max' = 0.
   Regresa el puntero removido de la lista en caso de sobrepasar el nmero
   de nodos 'max'.*/

void query_result_push_dup(query_result *qr, void *doc);

/* Igual pero sin importarle los duplicados, los acumula al primer nivel */
void query_result_push_no_dup(query_result *qr, void *doc);

/*Quita el primer elemento */
void query_result_pop_dup(query_result *qr);
void query_result_pop_no_dup(query_result *qr);

/* Mximo de todos los elementos "current" en los iteradores */
id_doc_t query_get_max(query *q);

/* Mnimo de todos los elementos "current" en los iteradores */
id_doc_t query_get_min(query *q);

#endif
