#ifndef _SKIP_LISTS_
#define _SKIP_LISTS_
#include <sys/types.h>
#include <super_malloc.h>

struct skip_node;
enum skip_node_type {HEADER, NODE, NIL};

struct skip_list {
  u_int8_t max_level; /* Mximo nivel para el nmero de elementos */
  u_int32_t num_nodes; /* Nmero de nodos utilizados actualmente */
  float prob; /* probabilidad*/
  /* Funcin de comparacin para los datos a - b
     a es siempre puesto al nuevo data
     b es siempre el dato que ya existe en la lista
     negativo si  b > a,
     positivo si a > b,
     cero     si a == b */
  int (*cmp)(void *a, void *b); 
  /* Funcin que maneja los datos duplicados, el primer argumento es el dato
     a insertar y el 2do el dato existente. Regresa el dato escogido. Debe
     liberar en caso necesario la memora no utilizada */
  void* (*dup)(void *a, void *b);
  /* Funcin para alojar memora, se llama si es necesario guardar el dato,
     si es null se utiliza el dato sin pasar por la funcin.
     No se llama si se encuentra un duplicado, en dup se debe alojar memora
     si es necesario. */
  void* (*alloc)(void *a);
  struct skip_node * head; /* Nodo vacio de cabeza */
  struct skip_node * nil; /* Nodo vacio final*/
};

struct skip_node {
  enum skip_node_type ntype; /* Tipo de nodo */
  /* Nivel del nodo */
  u_int8_t level;
  /* Dato */
  void *data;
  /* Arreglo de punteros siguientes */
  struct skip_node ** forward;
  /* El anterior */
  struct skip_node  * backward;
  /* Elemento que apunta el nodo al mximo nivel del nodo, util para remover
     por nodo y para manejo de duplicados */
  struct skip_node  * prev_max_level;
};

struct skip_list_iterator {
  struct skip_list *list; /* Lista iterada */
  struct skip_node *node; /* Nodo actual */
  u_int32_t position; /* Nmero de nodo (aproximado si hay multitarea)*/
};
typedef struct skip_node skip_node;
typedef struct skip_list skip_list;
typedef struct skip_list_iterator skip_list_iterator;


/* Crea un nuevo iterador para skip_list*/
int skip_list_iterator_new(skip_list *list, skip_list_iterator **itr);
/* Cierra un iterador */
void skip_list_iterator_close(skip_list_iterator *itr);
/* Regresa 1 si existen ms elementos inexplorados por el iterador */
int skip_list_iterator_has_next(skip_list_iterator *i);
/* Regresa el proximo elemento del iterador */
void *skip_list_iterator_next(skip_list_iterator *i);

#define skip_node_next(skipnode) ((skipnode)->forward[0])
#define skip_node_prev(skipnode) ((skipnode)->backward)
 
/* Crea un nuevo skip_node */
struct skip_node * skip_node_new(enum skip_node_type ntype, void *d, int level);

/* Inserta un nuevo skip node delante de ant con nivel: level*/
struct skip_node * skip_node_insert(struct skip_node * ant,
				    void *data, int level,
				    struct skip_node *nil);

/* Search a value */
void * skip_list_search(struct skip_list * list, void *value);
/* Libera un skip node*/
void skip_node_free(struct skip_node *);

/* Libera un skip_list, no libera datos, |f_free| es una funcin para
   liberar la memoria de los datos, si es NULL se debe usar
   skip_list_iterator_XXXXX con una funcin de liberacin de memoria
   para los datos antes de utilizar skip_list_free o no ser liberada la
   memoria */
void skip_list_free(struct skip_list *l, void (*f_free)(void *d));

/* Crea un skip_list con la probabilidad 'prob' y mximo nmero de elementos
   'max_num', 'cmp' Es la funcin de comparacin para insertar y borrar.
   |dup| maneja los duplicados, si es NULL se apilan los duplicados.
   |alloc| es llamada con data para producir un nuevo elemento al insertar,
   si el elemento existe y se manejan los duplicados se llama dup. En |dup|
   se debe alojar o borrar memora si es necesario. Si |alloc| es NULL
   se utiliza el mismo dato dado, sin pasar por funcin.
*/
struct skip_list * skip_list_new(float prob, int max_num,
				 int (*cmp)(void *obj, void *en_lista),
				 void *(*dup)(void *obj, void *en_lista),
				 void *(*alloc)(void *data));
				

/* Encuentra el elemento anterior al valor "value", usada en acess, insert,
   delete.
   'cmp' es la funcin de comparacin. El primer argumento es el puntero del
   elemento dentro de la lista, el segundo es elemento a comparar. El valor
   a regresar es int(a) - int(b).
  donde a es el valor de un elemento de skip_list */
struct skip_node * skip_list_find_prev_node(struct skip_list * list,
				  void *value, int (*cmp)(void *, void *));

/* Nivel aleatorio */
int skip_list_random_level(struct skip_list *);

#define skip_node_print(node, id)  \
fprintf(stderr, id " POINTER: %p, LEVEL: %d, TYPE: %d," \
 "DPOINTER: %p, DATA: %d\n", node, node->level, node->ntype, node->data, \
 ((node->data != NULL) ? *((int *)node->data) : -1))

#define skip_node_print_str(node, id) \
fprintf(stderr, id " POINTER: %p, LEVEL: %d, TYPE: %d," \
 "DPOINTER: %p, DATA: %s\n", node, node->level, node->ntype, node->data, \
	((node->data != NULL) ? (char *) (node->data) : "NULL"))

/* Inserta el dato 'data' en  'list' */
void skip_list_insert(struct skip_list *list, void *data);

/* Remueve el nodo cuyo dato es 'data' devuelve 'data' del elemento borrado,
 NULL si data no se encuentra */
void* skip_list_remove(struct skip_list *list, void *data);
/* remueve el nodo 's', devuelve 'data' de 's'*/
void* skip_list_remove_by_node(struct skip_list *list, struct skip_node *s);
/* remueve el ltimo nodo de la lista, devuelve 'data' del ltimo nodo */
/*void* skip_list_remove_last (struct skip_list *list);*/
/* remueve el ltimo nodo de la lista, devuelve 'data' del ltimo nodo */
void* skip_list_remove_first (struct skip_list *list);

/* Iteradores */

/* Visita hacia adelante los elementos de 'list'  con la funcin 'visit',
   visit debe regresar 1 a menos que se desee finalizar la iteracin, arg es
   el segundo argumento de 'visit' */

void skip_list_iterator_forward(struct skip_list *list,
				int (*visit) (skip_node *, void *), void *arg);

/* Igual a forward pero hace la visita al revs */
/*void skip_list_iterator_backward(struct skip_list *list,
  int (*visit)(skip_node *, void *), void *arg);*/
#endif /* _SKIP_LISTS_ */
