#!/usr/bin/python
#coding: utf-8;
############################################################################
#                                                                          #
#    This file 'distributor.py'                                            #
#    is part of 'SPyDI: Simple Python Distributed Indexing'                #
#    Copyright (C) 2004-2005 by Eric Sadit Tellez Avila                    #
#    sadit@lsc.fie.umich.mx or donsadit@gmail.com                          #
#                                                                          #
#    This program is free software; you can redistribute it and#or modify  #
#    it under the terms of the GNU General Public License as published by  #
#    the Free Software Foundation; either version 2 of the License, or     #
#    (at your option) any later version.                                   #
#                                                                          #
#    This program is distributed in the hope that it will be useful,       #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#    GNU General Public License for more details.                          #
#                                                                          #
#    You should have received a copy of the GNU General Public License     #
#    along with this program; if not, write to the                         #
#    Free Software Foundation, Inc.,                                       #
#    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             #
############################################################################

from threadutils import ThreadPool
import random
import SPyRO as spyro
import warn

class Dummy:
    def __init__(self, obj = None):
        self.obj = obj

    def findhost(self, key):
        """ Find a host related to |key| """
        return self.obj

    def unregister(self, host):
        """ Unregister a host """
        pass

    def registerbyaddress(self, addressport, location, objname,
                          auth=None, keys=None):
        """ Register an object by address """
        comm = spyro.TClient(addressport, location)
        self.register(comm.getobj(objname,auth), keys)
        warn.log("Registered %s with keys %s"%((addressport, location),
                 repr(keys)))

    def register(self, remote, keys):
        """ Register the |remote| object linked to the given |keys|"""
        self.obj = remote
    
    def execute(self, key, methodname, *args, **kwargs):
        """ Execute the method |methodname| with args and keyword arguments
        |kwargs| in the context of |key|. Executes in many hosts (as
        specified by the distributor algorithm)"""
        if self.obj is None: return None
        f = getattr(self.obj, methodname)
        return [ f(*args, **kwargs) ]

    def retrieve(self, key, methodname, *args, **kwargs):
        """ Exact to |execute| method but only executed in a single object """
        f = getattr(self.findhost(key), methodname)
        return f(*args, **kwargs)

    def objectlist(self, keys):
        """ Returns a list with the objects related to |keys|"""
        return [self.obj]

    def allobjects(self):
        """ Return all posible objects """
        return [self.obj]

    def close(self):
        """ Close the distributor """
        pass

class ByKey(Dummy):
    def __init__(self, poolSize = 4):
        self.dict = {}
        self.threadPool = ThreadPool(poolSize)
        self.__closed = False
    
    def findhost(self, key):
        v = self.dict[key]
        return v[random.randint(0, len(v) - 1)]
        
    def unregister(self, host):
        for k,v in self.dict.keys():
            try:
                v.remove(k)
            except ValueError:
                pass
    
    def register(self, remote, keys):
        for k in keys:
            a = self.dict.get(k, [])
            a.append(remote)
            self.dict[k] = a

    def __callback(self, res, exc, args):
        if exc is None: args[1].append(res)

    def __execute(self, values, methodname,  args, kwargs):
        exc = False
        res = []
        try:
            for v in values:
                f = getattr(v, methodname)
                self.threadPool.execute((self.__callback, res),
                                        f, *args, **kwargs)
        except Exception, err:
            exc = True
            warn.warn("Distributor ByKey", err)
        return res
    
    def execute(self, keys, methodname, *args, **kwargs):
        return self.__execute(self.objectlist(keys), methodname, args, kwargs)

    def objectlist(self, keys):
        v = []
        for k in keys:
            d = self.dict.get(k, None)
            if d: v.append(d[random.randint(0, len(v) - 1)])
        return v

    def allobjects(self):
        res = []
        for v in self.dict.values():
            res.extend(v)
        return res
    
    def close(self):
        self.__closed = True
        self.threadPool.close()

    def __del__(self):
        if not self.__closed:
            self.close()
            self.threadPool.close()

class Random(Dummy):
    def __init__(self, spread = False, poolSize = 2):
        self.pool = []
        self.spread = spread
        self.threadPool = ThreadPool(poolSize)
        self.__closed = False
    
    def findhost(self, key):
        return self.pool[random.randint(0, len(self.pool) - 1)]

    def unregister(self, obj):
        self.pool.remove(obj)
    
    def register(self, obj, keys):
        self.pool.append(obj)

    def __callback(self, res, exc, args):
        if exc is None:
            args[1].append(res)
        
    def execute(self, key, methodname, *args, **kwargs):
        res = []
        exc = False
        thrpool = self.threadPool
        if self.spread or key is None:
            #thrpool.begin(len(self.pool))
            #print len(self.pool)
            try:
                #thrpool.begin(len(self.pool) - 1)
                for v in self.pool:
                    f = getattr(v, methodname)
                    #warn.debug("KEY A:",key)
                    thrpool.execute((self.__callback, res),
                                   f, *args, **kwargs)
                    #res.append(f(*args, **kwargs))
                    #warn.debug("KEY B:",key)
                #thrpool.end()
            except Exception, err:
                exc = True
                warn.warn("Random distributor (spread behavior) ",err)
        else:
            try:
                f = getattr(self.findhost(None), methodname)
                res.append(f(*args, **kwargs))
            except Exception, err:
                exc = True
                warn.warn("Random distributor (random behavior) ", err)
        return res

    def close(self):
        self.__closed = True
        self.threadPool.close()

    def objectlist(self, keys):
        if self.spread or keys is None:
            return self.pool
        else:
            return [self.findhost(keys)]

    def allobjects(self):
        return self.pool

    def __del__(self):
        if not self.__closed:
            self.close()

# class Hub:
#     def __init__(self, __dist):
#         self.__dist = __dist

#     def __getattr__(self, method):
#         return lambda *args, **kwargs: self.__dist.execute(None, methodname=method,*args,**kwargs)

#if __name__ == "__main__":
#     for u in (("lsc.fie.umich.mx",8000,"/SPyRO"),
#               ("lsc.fie.umich.um",8110,"/SPyRO"),
#               ("lsc.fie.umich.um",8110,"/SPyRO"),
#               ("lsc.fie.xmmso.um",110,"/SPyRO")):
#         b.register(u, ('some', 'key', 'words'))

#     for i in ('x' , 'some', 'key', 'words'):
#         try:
#             print b.findhost(i)
#         except:
#             warn.debug("Does not exists '%s'"%(i))
#    b.close()
