#!/usr/bin/python
#coding: utf8;

############################################################################
#                                                                          #
#    This file 'spydex.py'
#    is part of 'SPyDI: Simple Python Distributed Indexing                 #
#    Copyright (C) 2004-2005 by Eric Sadit Téllez Avila                    #
#    sadit@lsc.fie.umich.mx or donsadit@gmail.com                          #
#                                                                          #
#    This program is free software; you can redistribute it and#or modify  #
#    it under the terms of the GNU General Public License as published by  #
#    the Free Software Foundation; either version 2 of the License, or     #
#    (at your option) any later version.                                   #
#                                                                          #
#    This program is distributed in the hope that it will be useful,       #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#    GNU General Public License for more details.                          #
#                                                                          #
#    You should have received a copy of the GNU General Public License     #
#    along with this program; if not, write to the                         #
#    Free Software Foundation, Inc.,                                       #
#    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             #
############################################################################

__doc__ = """This is the replacement of SPyDI module (called spydex).
spydex will be completed and tested as thread safe then it will be
called SPyDI

spydex is a fast fulltext indexer written in python, using the bsddb module.

We also use psyco and pyrex to increase the processing speed, and computing.

"""

import sre
import sys, os
from threadutils import ThreadPool
import bsddb
_DB_SET_RANGE = bsddb.db.DB_SET_RANGE
_DB_SET = bsddb.db.DB_SET
from struct import pack, unpack
import threading
import warn
from math import log, sqrt

_TF_ACCUR = 100000.0
_SEP_STRING = "\000"
_remove_zeros = sre.compile("^\000+")
_keywordregex = sre.compile(r"[\w:/!\.][\w:/\.]{0,63}",sre.L)

def _cmp_2dkey(x,y):
    return cmp(x[1],y[1])

def _cmp_gdata_(ga,gb):
    return cmp(ga[1][2],gb[1][2])

def _xtract2(x):
    return x[1]

class AlreadyIndexedVector(Exception): pass

class VectorDocument:
    """ Represent a vectorized document """
    def __init__(self, maxfreq = 0, vector = None):
        """ Initialize the VectorDocument
        |maxfreq| The maximum frecuency in the document
        |vector| The vector of terms (dict representation)"""
        if vector is None:
            self.maxfreq = 0
            self.vector = {}
        else:
            self.maxfreq = maxfreq
            self.vector = vector

class Query:
    def __init__(self, indexer, vectorquery, vectorkeywords, gdatalist):
        """ Creates a query object with
        |indexer| The indexer used to resolve queries
        |vectorquery| Vector of the query, a VectorDocument object
        |gdatalist| Word statistics  """
        self.indexer = indexer
        self.vectorquery = vectorquery
        self.vectorkeywords = vectorkeywords
        self.gdatalist = gdatalist
        self.cursors_wfreq = {}
        self.cursors_wdocid = {}
        self.totaldocuments = float(len(self.indexer.filedata)+1)
        self.__closed = False
        
        
    def getCollectionSize(self):
        """ Gets the collection size """
        return self.totaldocuments

    def setCollectionSize(self, newsize):
        """ Set the collection size, virtually """
        self.totaldocuments = newsize

    def getgdata(self):
        """ Get gdata  """
        return self.gdatalist

    def setgdata(self,gdatanew):
        """ Set gdata, does not validate gdatanew """
        gdatanew.sort(_cmp_gdata_)
        self.gdatalist = gdatanew
    
    def __permutation(self, x):
        """ Retrieves and replace the name of the docid (first field) """
        return self.indexer.get_urlfromdocid(x[0]), x[1]

    def __del__(self):
        if not self.__closed: self.close()
    
    def init(self, initall = False):
        pass

    def _init_cursors(self, wlist, setcursors):
        cursors = {}
        for word,gdata in self.gdatalist:
            if gdata[2] == 0:  # in other words numdocs == 0
                pass
            else:
                cursor = wlist.cursor()
                if setcursors: cursor.get(word, _DB_SET_RANGE)
                cursors[word] = cursor
        print repr(self.gdatalist)
        return cursors

    def _init_cursors_wfreq(self, setcursor):
        self.cursors_wfreq=self._init_cursors(self.indexer.invlistwfreq,
                                             setcursor)

    def _init_cursors_wdocid(self, setcursor):
        self.cursors_wdocid=self._init_cursors(self.indexer.invlistwdocid,
                                              setcursor)
                                             
    def close(self):
        if self.__closed: return
        self.__closed = True
        for c in self.cursors_wfreq.values(): c.close()
        for c in self.cursors_wdocid.values(): c.close()

    def doquery(self, begin, count, qtype):
        if qtype == "rank_vector":
            return self.rank_vector(begin, count)
        elif qtype == "rank_freq":
            return self.rank_freq(begin, count)
        elif qtype == "rank_fastvector":
            return self.rank_fastvector(begin, count)
        else:
            return self.rank_vector(begin, count)
    
    def rank_freq(self, begin, count):
        """ A simple ranking function based on a TFxIDF scheme, but without
        the TF x IDF :S"""
        self._init_cursors_wfreq(True)
        total = begin+count
        gdatalist = self.gdatalist
        _loads_wfreq = self.indexer.loads_wfreq
        lres = {}
        cursors = self.cursors_wfreq
        _run_valid_docid = not (self.vectorkeywords is None)
        for word, gdata in gdatalist:
            wordmaxfreq, maxfreq, numdocs = gdata
            if numdocs == 0: continue
            cur = cursors[word]
            k = 0
            record = cur.current()
            while record and k < numdocs and k < total:
                k = k + 1
                _word, url, freq = _loads_wfreq(record)
                if _run_valid_docid and not self.valid_docid(url):
                    record = cur.next()
                    continue
                docf = lres.get(url,0)
                lres[url] = freq + docf
                record = cur.next()
            if len(lres) >= total: break
        xres = map(self.__permutation, lres.items())
        xres.sort(_cmp_2dkey, reverse=True)
        return xres[int(begin):int(total)]

    def rank_fastvector(self, begin, count):
        """ A modified (to be fast and good, but not exact) vectorial
        ranking function """
        self._init_cursors_wfreq(True)
        gdatalist = self.gdatalist
        total = begin+count
        _loads_wfreq = self.indexer.loads_wfreq
        _loads_urldata = self.indexer._loads_urldata
        _get_urldata = self.indexer.filedata.get
        lres = {}
        qTFIDF = []
        qDEN = 0
        N = float(self.totaldocuments)
        maxf = float(self.vectorquery.maxfreq)
        vector = self.vectorquery.vector
        # Query vector processing step
        for term, gdata in gdatalist:
            numdocs = gdata[-1]
            #if numdocs == 0: continue
            maxfreq = vector[term]
            #The zero records may be ignored
            if numdocs == 0:
                tfidf = 1
            else:
                tfidf = (maxfreq / maxf) * log(N / numdocs)
            qTFIDF.append(tfidf)
            qDEN = qDEN + tfidf
        qDEN = sqrt(qDEN)
        # vector collection processing step
        mapdocdata = {}
        wordindex = -1
        _run_valid_docid = not (self.vectorkeywords is None)
        cursors = self.cursors_wfreq
        for word, gdata in gdatalist:
            wordmaxfreq, accdata, numdocs = gdata
            wordindex = wordindex + 1
            if numdocs == 0: continue
            cur = cursors[word]
            k = 0
            record = cur.current()
            _maxtotal = total * 2
            while record and k < numdocs and len(mapdocdata) < _maxtotal:
                _word, docid, freq = _loads_wfreq(record)
                k = k + 1
                if _run_valid_docid and not self.valid_docid(docid):
                    record = cur.next()
                    continue
                docdata = mapdocdata.get(docid)
                if docdata is None:
                    # List data: [docmaxfreq, dTFIDF*qTFIDF, dDEN,
                    # doctermsinquery]
                    docdata = [_loads_urldata(_get_urldata(docid)), 0.0, 0.0,
                               0.0]
                    mapdocdata[docid] = docdata
                if numdocs == 0:
                    dTFIDF = 0.00001
                else:
                    #dTFIDF = (float(freq) / docdata[0][0]) * log(N / numdocs)
                    dTFIDF = freq * log(N / numdocs)
                #print repr(docdata), qTFIDF
                docdata[1] = docdata[1] + dTFIDF * qTFIDF[wordindex]
                docdata[2] = docdata[2] + dTFIDF * dTFIDF
                docdata[3] = docdata[3] + freq
                record = cur.next()
            #if len(lres) >= total: break
        # Dot product
        #wordindex = 0
        lq = len(self.vectorquery.vector)
        if len(gdatalist) > 0:
            maxnumdocs = gdatalist[-1][1][2]
        else:
            maxnumdocs = 1
        for docid, docdata in mapdocdata.items():
#            lres[docid] = docdata[1] / (sqrt(docdata[2]) * qDEN)
            docmaxfreq, docdiffterms, doctotalterms = docdata[0]
            compfac=((doctotalterms-docdata[3])/docdiffterms / docmaxfreq) # TF
            #TFxIDF(log(log(N))=magica)xNveces
            compfac=compfac * log(N/maxnumdocs) * max(0, (docdiffterms - lq))
            lres[docid] = docdata[1] / (sqrt(docdata[2] + compfac) * qDEN)
            #wordindex = wordindex + 1
        # Sorting step
        xres = map(self.__permutation, lres.items())
        xres.sort(_cmp_2dkey, reverse=True)
        return xres[int(begin):int(total)]

    def valid_docid(self, docid_leader, keywords = None):
        """ Probes a docid to be valid in the keyword universe
        Using the boolean model, The negation of a keyword is performed
        with the |!| character following the keyword. """
        #keywords = ('!soda','stereo')
        print "v000"
        cursor = self.indexer.invlistwdocid.cursor()
        #print "*************************************>", docid_leader
        if keywords is None: keywords = self.vectorkeywords.vector.keys()
        for word in keywords:
            if word[0] == '!':
                word = word[1:]
                musthave = False
            else:
                musthave = True
            #key = word + _SEP_STRING + docid_leader
            key = word + _SEP_STRING + docid_leader
            print "v111A", repr(key)
            #record = cursor.get(key, _DB_SET_RANGE)
            record = cursor.get(key, _DB_SET)
            print "v111B", repr(key)
            if musthave:
                if key != record[0]: return False
            else:
                if key == record[0]: return False
        return True
    
    def rank_vector(self, begin, count): 
        """ A modified (to be fast (but not so fast), but good, but not exact)
        vectorial ranking function """
        #self._init_cursors()
        self._init_cursors_wdocid(False)
        wlist = self.indexer.invlistwfreq
        gdatalist = self.gdatalist
        total = begin+count
        _loads_wfreq = self.indexer.loads_wfreq
        _loads_wdocid = self.indexer.loads_wdocid
        _loads_urldata = self.indexer._loads_urldata
        _get_urldata = self.indexer.filedata.get
        _run_valid_docid = not (self.vectorkeywords is None)
        lres = {}
        qTFIDF = []
        qDEN = 0
        N = float(self.totaldocuments)
        maxf = float(self.vectorquery.maxfreq)
        vector = self.vectorquery.vector
        # Query vector processing step
        for term, gdata in gdatalist:
            numdocs = gdata[-1]
            #if numdocs == 0: continue
            maxfreq = vector[term]
            #The zero records may be ignored
            if numdocs == 0:
                tfidf = 1
            else:
                tfidf = (maxfreq / maxf) * log(N / numdocs)
            qTFIDF.append(tfidf)
            qDEN = qDEN + tfidf
        qDEN = sqrt(qDEN)
        # vector collection processing step
        mapdocdata = {}
        wordindex = -1
        cursors = self.cursors_wdocid
        cursor_leader = None
        record_leader = None
        for word, gdata in gdatalist:
            wordmaxfreq, accdata, numdocs_leader = gdata
            wordindex = wordindex + 1
            if numdocs_leader == 0: continue
            cursor_leader = wlist.cursor()
            record_leader = cursor_leader.get(word, _DB_SET_RANGE)
            break
        print "********1111111--------xxxxxxxxxxxx",record_leader
        #ngdatalist = gdatalist[wordindex:]
        ngdatalist = gdatalist
        k = 0
        _maxtotal = total * 2
        while record_leader and k < numdocs_leader and len(mapdocdata) < _maxtotal:
            print "********1111111A"
            word_leader, docid_leader, freq_leader=_loads_wfreq(record_leader)
            print "********1111111B",_run_valid_docid
            k = k + 1
            if _run_valid_docid and not self.valid_docid(docid_leader):
                print "********1111111C"
                record_leader = cursor_leader.next()
                print "********1111111D"
                continue
            print "********1111111E"
            wordindex = -1
            # List data: [docmaxfreq, dTFIDF*qTFIDF, dDEN, doctermsinquery]
            docdata = [_loads_urldata(_get_urldata(docid_leader)),
                       0.0, 0.0, 0.0]
            for word, gdata in ngdatalist:
                wordmaxfreq, accdata, numdocs = gdata
                wordindex = wordindex + 1
                try:
                    cursor = cursors[word]
                except KeyError:
                    continue
                key = word + _SEP_STRING + docid_leader
                record = cursors[word].get(key, _DB_SET_RANGE)
                _word, docid, freq = _loads_wdocid(record)
                if numdocs == 0 or key != record[0]:
                    dTFIDF = 0.00001
                else:
                    #dTFIDF = (float(freq) / docdata[0][0]) * log(N / numdocs)
                    dTFIDF = freq * log(N / numdocs)
                docdata[1] = docdata[1] + dTFIDF * qTFIDF[wordindex]
                docdata[2] = docdata[2] + dTFIDF * dTFIDF
                docdata[3] = docdata[3] + freq
            record_leader = cursor_leader.next()
            mapdocdata[docid_leader] = docdata
        # Dot product
        #wordindex = 0
        lq = len(self.vectorquery.vector)
        # the approximated number of documents
        if len(gdatalist) > 0:
            maxnumdocs = gdatalist[-1][1][2]
        else:
            maxnumdocs = 1
        #print "---------->",maxnumdocs, N
        for docid, docdata in mapdocdata.items():
            docmaxfreq, docdiffterms, doctotalterms = docdata[0]
            #TFxIDF(log(log(N))=magica)xNveces
            compfac=((doctotalterms-docdata[3]) / docdiffterms / docmaxfreq )
            #compfac= compfac * log(N/docmaxfreq) * max(0, (docdiffterms - lq))
            compfac= compfac * log(N / maxnumdocs) * max(0,(docdiffterms - lq))
            #print "XXXXXXXXXXXXXXXXXXXXXXXx>\n", repr(docdata), compfac, qDEN
            #print "YYYYYYYYYYYYYYYYYYYYYYYY>", N, docmaxfreq
            lres[docid] = docdata[1] / (sqrt(docdata[2] + compfac) * qDEN)
            #wordindex = wordindex + 1
            #####lres[docid] = docdata[1] / (sqrt(docdata[2]) * qDEN)
        # Sorting step
        xres = map(self.__permutation, lres.items())
        xres.sort(_cmp_2dkey, reverse=True)
        #print "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx SPIDEX:", begin, total
        return xres[int(begin):int(total)]

translationtable='\x00\x01\x02\x03\x04\x05\x06\x07\x08\t\n\x0b\x0c\r\x0e\x0f\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b\x1c\x1d\x1e\x1f !"#$%&\'()*+,-./0123456789:;<=>?@abcdefghijklmnopqrstuvwxyz[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~\x7f\x80\x81\x82\x83\x84\x85\x86\x87\x88\x89\x8a\x8b\x8c\x8d\x8e\x8f\x90\x91\x92\x93\x94\x95\x96\x97\x98\x99\x9a\x9b\x9c\x9d\x9e\x9f\xa0\xa1\xa2\xa3\xa4\xa5\xa6\xa7\xa8\xa9\xaa\xab\xac\xad\xae\xaf\xb0\xb1\xb2\xb3\xb4\xb5\xb6\xb7\xb8\xb9\xba\xbb\xbc\xbd\xbe\xbf\x61\x61\x61\x61\x61\x61\x61\xc7\x65\x65\x65\x65\x69\x69\x69\x69\xd0\x6e\x6f\x6f\x6f\x6f\x6f\xd7\x6f\x75\x75\x75\x75\xdd\xde\xdf\x61\x61\x61\x61\x61\x61\xe6\xe7\x65\x65\x65\x65\x69\x69\x69\x69\xf0\x6e\x6f\x6f\x6f\x6f\x6f\xf7\x6f\x75\x75\x75\x75\xfd\xfe\xff'

def normalize(word):
    """ Apply the translation of |translationtable| to |word| """
    global translationtable
    return word.translate(translationtable)
    
class FulltextIndexer:
    """ The main class """
    def __init__(self, basename = "",
                 cachesize = 32 << 20, mode = 0644,
                 wordreg = None, normalizer = None,
                 stemmer = None, stopwords = None):
        """ The fulltext indexer
        |basename| is the basename to create the text databases
        |cachesize| Size of the memory cache (greater is better)
        |mode| Mode of created databases (permission access)
        |wordreg| Is the regular expression to read words, If is None
        letter and digits and the word length must be between 1 and 16
        characters
        |normalizer| Is the word and character normalizer, is applied before
        the stemmer. The normalizer is a function that receives every word
        (str) and returns a normalized word (str). If None the normalize
        function in the module is used.
        |stemmer| Is for every word to create stemmed thesaurous.
        The default is None
        |stopwords| A dict with the stopwords (will not be indexed). It must
        be stemmed because they are compared with the stemmed words  """
        db = bsddb._db
        mode = 0644
        self.stemmer = stemmer
        self.stopwords = stopwords
        self.__closed = False
        if wordreg is None: wordreg = sre.compile(r"[\d\w]{1,16}",sre.L)
        self.wordreg = wordreg
        if normalizer is None: normalizer = normalize
        self.normalizer = normalizer
        self.dbEnvIndex = db.DBEnv()
        self.dbEnvIndex.set_cachesize(0, cachesize, 0)
        envflags = ( db.DB_INIT_MPOOL |
                     #db.DB_INIT_CDB |
                     db.DB_INIT_LOG |
                     db.DB_INIT_TXN |
                     db.DB_INIT_LOCK |
                     db.DB_CREATE | db.DB_PRIVATE | db.DB_THREAD )
        self.dbEnvIndex.open(basename, envflags)
        dbflags = db.DB_THREAD|db.DB_CREATE
        self.index=db.DB(self.dbEnvIndex, 0)
        self.index.open("thesaurus.db",
                        #dbname="thesaurus",
                        dbtype=db.DB_BTREE,
                        flags=dbflags,
                        mode=mode)
        self.invlistwfreq=db.DB(self.dbEnvIndex, 0)
        self.invlistwfreq.set_flags(db.DB_DUP | db.DB_DUPSORT)
        self.invlistwfreq.open("invlistwfreq.db",
                           #dbname="invlistwfreq",
                           dbtype=db.DB_BTREE,
                           flags=dbflags,
                           mode=mode)
        self.invlistwdocid=db.DB(self.dbEnvIndex, 0)
        self.invlistwdocid.set_flags(db.DB_DUP | db.DB_DUPSORT)
        self.invlistwdocid.open("invlistdocid.db",
                           #dbname="invlistwfreq",
                           dbtype=db.DB_BTREE,
                           flags=dbflags,
                           mode=mode)
        self.filedata=db.DB(self.dbEnvIndex, 0)
        self.filedata.open("filedata.db",
                           dbname="filedata",
                           dbtype=db.DB_BTREE,
                           flags=dbflags,
                           mode=mode)

        self.url2docid=db.DB(self.dbEnvIndex, 0)
        self.url2docid.open("filedata.db",
                        dbname="url2docid",
                        dbtype=db.DB_BTREE,
                        flags=dbflags,
                        mode=mode)
        self.docid2url=db.DB(self.dbEnvIndex, 0)
        self.docid2url.open("filedata.db",
                        dbname="docid2url",
                        dbtype=db.DB_RECNO,
                        flags=dbflags,
                        mode=mode)
        self.lock = threading.Lock()

    def __del__(self):
        if not self.__closed: self.close()

    def close(self):
        if self.__closed: return
        self.__closed = True
        self.index.close()
        self.invlistwfreq.close()
        self.invlistwdocid.close()
        self.filedata.close()
        self.docid2url.close()
        self.url2docid.close()
        self.dbEnvIndex.close()

    def sync(self):
        pass

    def vectorizechunkedandappend(self, url, kwargs_vector, vecdoc=None):
        """ Do a vectorizechunked and append the vector. return None """
        vecdoc = self.vectorizechunked(kwargs_vector, vecdoc)
        self.appendvector(url, vecdoc)
        
    def vectorizechunked(self, kwargs_vector, vecdoc=None):
        """ A chunked version of vectorize.
        |kwargs_vector| is a sequence of dictionaries with
        the arguments to be passed to vectorize
        |datafile| must be named as 'data' and vecdoc is internal to
        vectorizechunked
        """
        for kwargs in kwargs_vector:
            data = kwargs.pop('data')
            vecdoc = self.vectorize(data,vecdoc,**kwargs)
        return vecdoc

    
    def vectorize(self, datafile, vecdoc = None,
                  wordreg = None, normalizer = None,
                  stemmer = None, stopwords = None):
        """ Vectorize a document
        |datafile| Is the documents data
        |vecdoc| The VectorDocument. if |vecdoc| is None a new one is
        created, if |vecdoc| is not None then append to vecdoc
        the |datafile| vector. Useful to chunked vectorizations.
        |wordreg| A regular expression to read words, if None the
        indexer wordreg will be used
        |normalizer| A functions that normalize every word, if None
        the indexer wordreg will be used
        |stemmer| Is for every word to create stemmed thesaurous.
        If None is given, the indexer stemmer is used
        |stopwords| A dict with the stopwords (words that must not be
        indexed). If None is given, the indexer stopwords are used.
        Note1: The stopwords must be already stemmed
        Note2: If you need to disable any inheretable argument as
        wordreg,normalizer,stemmer, or stopwords just set to False,
        do not use None. """
        maxfreq = 0
        if stemmer is None: stemmer = self.stemmer
        if stopwords is None: stopwords = self.stopwords
        if vecdoc is None:
            vecdoc = VectorDocument()
            dicdoc = vecdoc.vector
        if wordreg is None:
            if type(wordreg) is str: wordreg = sre.compile(wordreg)
            wordreg = self.wordreg
        if normalizer is None: normalizer = self.normalizer
        for word in wordreg.finditer(datafile):
            word = word.group()
            if normalizer:
                word = normalizer(word)
            if stemmer: word = stemmer(word)
            if stopwords and stopwords.has_key(word): continue
            freq = dicdoc.get(word, 0) + 1
            maxfreq = max(freq, maxfreq)
            dicdoc[word] = freq
        vecdoc.maxfreq = maxfreq
        return vecdoc

    def get_docidfromurl(self, filename):
        """ Get the docid of the url name |filename|"""
        self.lock.acquire()
        docid = self.url2docid.get(filename)
        if docid is None:
            docid = self.docid2url.append(filename)
            #docid = _remove_zeros.sub("",pack(">I",docid))
            docid = pack(">I",docid)
            self.url2docid.put(filename, docid)
        self.lock.release()
        return docid

    def get_urlfromdocid(self, docid):
        """ Get the url associated to |docid|"""
        return self.docid2url.get(int(unpack(">I",docid)[0]))

    def _loads_freq(self, wordfreq):
        #return ~unpack(">i",wordfreq)[0]
        return unpack(">i", wordfreq)[0] / _TF_ACCUR

    def _dumps_freq(self, freq):
        #return pack(">i", ~freq)
        return pack(">i", int(_TF_ACCUR * freq))
    
    def dumps_wfreq(self, word, docid, freq):
        """ packs the word, docid and freq and returns
        the key and data of the invlistwfreq text database """
        #return word, docid + self._dumps_freq(freq) # Experimental, sort by doc
        return word, self._dumps_freq(freq) + docid # Experimental, sort by freq
        #return word + _SEP_STRING + self._dumps_freq(freq), docid # Stable

    def loads_wfreq(self, record):
        """ returns the word, docid, freq. See dumps_wfreq """
        #return record[0][:-5], record[1], self._loads_freq(record[0][-4:])
        return record[0], record[1][-4:], self._loads_freq(record[1][:4])# Experimental, s

    def dumps_wdocid(self, word, docid, freq):
        """ packs the word, docid and freq and returns
        the key and data of the invlistwfreq text database """
        #return word + _SEP_STRING +  docid, self._dumps_freq(freq)
        return docid, word + _SEP_STRING + self._dumps_freq(freq)


    def loads_wdocid(self, record):
        """ returns the word, docid, freq. See dumps_wfreq """
        #return record[0][:-5], record[0][-4:], self._loads_freq(record[1])
        return record[1][:-5], record[0], self._loads_freq(record[1][-4:])
    
    def _loads_urldata(self, urldata):
        return unpack("HHH", urldata)
    
    def _dumps_urldata(self, maxfreq, diffterms, totalterms):
        """ Dumps urldata
        |maxfreq| Maximum frequency in the document
        |diffterms| Number of different terms in the document
        |totalterms| Total number of terms in the document """
        return pack("HHH", maxfreq, diffterms, totalterms)

    def _loads_gdata(self, gdata):
        if gdata is None:
            return (0,0,0)
        else:
            return unpack("HII",gdata)

    def _dumps_gdata(self, wordmaxfreq, accfreq, numdocs):
        """ Dumps:
        wordmaxfreq. Maximum frecuency of this term in all documents
        accfreq. Accumulated frecuency of this term
        numdocs. The number of documents with this word """
        return pack("HII",wordmaxfreq, accfreq, numdocs)

    def appendvector(self, url, vecdoc):
        """ Append a new VectorDocument to the index """
        index = self.index
        invlistwfreq = self.invlistwfreq
        invlistwdocid = self.invlistwdocid
        lock = self.lock
        docid = self.get_docidfromurl(url)
        maxfreq = float(vecdoc.maxfreq)
        if maxfreq == 0: return
        totalterms = 0
        for word, freq in vecdoc.vector.items():
            totalterms = totalterms + 1            
            try:
                #lock.acquire()
                tf = freq / maxfreq
                key, data = self.dumps_wfreq(word, docid, tf)
                invlistwfreq.put(key,data)
                key, data = self.dumps_wdocid(word, docid, tf)
                invlistwdocid.put(key,data)
                #lock.release()
            except bsddb.db.DBKeyExistError:
                raise AlreadyIndexedVector, "Trying to reinsert a document, go away (for now)"
                #lock.release()
                #warn.warn("Trying to reinsert a document, go away (for now)")
                return
            wdata = index.get(word)
            if wdata is None:
                wordmaxfreq = freq
                wordfreq = freq
                numdocs = 1
            else:
                wordmaxfreq, wordfreq, numdocs = self._loads_gdata(wdata)
                numdocs = numdocs + 1
                wordmaxfreq = max(wordmaxfreq, freq)
                wordfreq = wordfreq + freq
            index.put(word, self._dumps_gdata(wordmaxfreq,wordfreq,numdocs))
        urldata = self._dumps_urldata(maxfreq,len(vecdoc.vector),totalterms)
        self.filedata.put(docid, urldata)

    def search(self, vectorquery, begin, count, qtype="rank_vector",
               vectorkeywords = None):
        q = self.newQuery(vectorquery, vectorkeywords)
        q.init()
        gdata = q.getgdata()
        results = q.doquery(begin, count, qtype)
        q.close()
        return gdata, results

    def parsequery(self, vectorquery, regexp = None):
        if type(vectorquery) in (tuple, list):
            return self.vectorize(" ".join(vectorquery),
                                  wordreg=regexp)
        elif type(vectorquery) is str:
            return self.vectorize(vectorquery,
                                  wordreg=regexp)
        return vectorquery
    
    def newQuery(self, vectorquery, vectorkeyword=None):
        """ Creates a new query from |vectorquery| query vector """
        index = self.index
        gdatalist = []
        vectorquery = self.parsequery(vectorquery, None)
        vectorkeyword = self.parsequery(vectorkeyword, _keywordregex)
        for term,freq in vectorquery.vector.items():
            gdatalist.append((term, self._loads_gdata(index.get(term))))
        gdatalist.sort(_cmp_gdata_)
        return Query(self, vectorquery, vectorkeyword, gdatalist)

class spydex(FulltextIndexer):
    def _setPool(self, poolSize):
        self.poolSize = poolSize
        self.threadPool = ThreadPool(poolSize)

    def __init__(self, poolSize=2, distrepository=None, textfilter=None,
                 **kwargs):
        FulltextIndexer.__init__(self, **kwargs)
        self.__closed = False
        self.replicas = 1
        self.distrepository = distrepository
        self.textfilter = textfilter
        self._setPool(poolSize)
    
    def indexFromRepositoryList(self, urllist):
        for url in urllist:
            try:
                self.indexFromRepository(url)
            except:
                warn.warn("indexFromRepository:", url)
    
    def indexFromRepository(self, url):
        """ Index something from the repository """
        site, string = self.distrepository.retrieve(url, "getandopen", url)
        vecdoc = self.vectorize(string)
        self.appendvector(url, vecdoc)
        veckey = self.vectorize(" ".join(site.keywords),
                                wordreg = sre.compile(r"[^\s]+"))
        veckey.maxfreq = vecdoc.maxfreq
        self.appendvector(url, veckey)

    def indexString(self, name, string, doc_id=None):
        """ Index a string """
        vecdoc = self.vectorize(string)
        self.appendvector(name, vecdoc)

    def indexMagicFile(self, path, doc_id=None):
        """ Index a file trying to guess useful structure of the file """
        print "Not implemented... yet!"

    def indexMagicFiles(self, paths):
        """ Index a file trying to guess useful structure of the file """
        for path in paths: self.indexMagicFile(path)

    def indexPlainFiles(self, paths):
        for path in paths: self.indexPlainFile(path)

    def indexPlainFile(self, path, doc_id=None, bufflen=4096):
        """ Index a plain text file"""
        f = file(path)
        vecdoc = self.vectorize(f.read())
        self.appendvector(path, vecdoc)
        f.close()
    
    def indexXMLFile(self, path, doc_id=None):
        """Index a XML file. Using its structure. """
        print "En cola XML File", path
        #self.threadPool.execute(None, self.__indexXML__, path)
        oai = citeseer.CiteseerOAI(self)
        self.indexString(path, oai.insert(path, path), doc_id)
        print "Termino XML File", path

    def indexXMLFiles(self, paths):
        for path in paths: self.indexXMLFile(path)

    def __del__(self):
        if not self.__closed:
            self.close()

    def close(self):
        self.__closed = True
        self.threadPool.close()
        FulltextIndexer.close(self)


#############################################################################
#####
#####   Test
#####
#############################################################################

import threading
numdocs = 0
processedbytes = 0
lock_index = threading.Lock()

def pruebaindexdocument(index, filename):
    global numdocs, processedbytes, lock_index
    try:
        datafile = file(filename)
    except IOError, err:
        warn.debug(err,": ",filename)
        return
    data = datafile.read()
    processedbytes = processedbytes + len(data)
    vecdoc = index.vectorize(data)
    lock_index.acquire()
    index.appendvector(filename, vecdoc)
    lock_index.release()
    numdocs = numdocs + 1
    sys.stdout.write("Numdocs: %s, Filename: %s, processed bytes: %s\n"%(
        numdocs, filename, processedbytes))
    datafile.close()

def prueba():
    global numdocs, processedbytes
    #index = FulltextIndexer(basename="datafiles/")
    index = spydex(basename="datafiles/",cachesize=128<<20)
    #pool = ThreadPool(2)
    #pool.begin(400)
    if len(sys.argv) < 2:
        warn.debug("Usage: %s list-of-files " % sys.argv)
        sys.exit(1)
    f = file(sys.argv[1])
    lines = f.readlines()#
    for filename in lines[:4000]:
         filename = filename.rstrip()
         #pool.execute(None, pruebaindexdocument, index, filename)
         pruebaindexdocument(index, filename)
         #index.indexPlainFile(filename)
    print "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX SOL"
    #pool.end()
    #query = file("/home/sadit/xml/oai_citeseer100001-57.xml").read()
    #queryfile = lines[len(lines[-1])].strip()
    #print "Similar to ", queryfile
    #queryfile = "/home/sadit/tmp/comp/www.atame.org/s/soda_stereo/angel_electrico-3.txt"
    #query = file(queryfile).read()
    query = "Soda Stereo angel electrico"
    #query = "Sufre Mamon"
    #query = "geometry geometry geometry line first search string a"
    vecquery = index.vectorize(query)
#     for key, values in index.invlistwdocid.items():
#         print "="  * 20
#         print repr(key), len(values)
#         print repr(values)
    index.close()
    sys.exit(0)
    print "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"
    gdata, results = index.search(vecquery,0,10,"rank_vector",'soda stereo dejarme')
    #gdata, results = index.search(vecquery,0,10,"rank_vector")
    #gdata, results = (), ()
    #print "**** GDATA: ", gdata
    print "**** FILES: "
    k = 0
    print "%s\t%s\t%s"%("Pos", "Rank", "Files")
    print "%s\t%s\t%s"%(("-----",)*3)
    for rank, url in results:
        print "%s\t%s\t%s"%(k, rank, url)
        k = k + 1
    sys.stdout.write("ndocs: %s, bytes: %s\n"%(numdocs, processedbytes))
    f.close()
    #pool.close()
    print threading.enumerate()

def test():
    import SPyRO as spyro
    import Repository
    addr=('0.0.0.0',6969)
    import socket
    try:
        import psyco
        psyco.profile()
    except:
        warn.debug("The psyco module is not present")

    ConnectionBroken = True
    numports = 10 
    while ConnectionBroken and numports:
        try:
            server=spyro.CommunicationServer(addr, '/SPyRO?PICKLE2')
        except socket.error, e:
            warn.warn(e)
            addr = (addr[0], addr[1]+numports)
            numports = numports - 1
        else:
            ConnectionBroken = False

    #basename = 'datafiles/' + str(numports) # para pruebas
    basename = 'datafiles/'
    index = spydex(basename=basename, poolSize=4)
    #atexit.register(index.close)
    server.setobj('index',index)
    #comm=spyro.CommunicationClient(('127.0.0.1',7777),'/SPyRO?PICKLE2')
    comm=spyro.TClient(('127.0.0.1',3001),'/SPyRO?PICKLE2',local_server=server)
    idxcontrol = comm.getobj('index')
    index.distrepository = comm.getobj("RepositoryDistributor") 
    print "regindexer"
    idxcontrol.regindexer(index)
    comm.close()
    try:
        print "Sirviendo en ", addr
        server.serve_forever()
    except Exception, e:
        warn.warn(e)
    server.close()
    index.close()


if __name__ == '__main__':
    import psyco
    psyco.profile()
    #test()
    prueba()
