class FulltextIndexer:
    """ The main class """
    def __init__(self, basename = "",
                 cachesize = 64 << 20, mode = 0644,
                 wordreg = None, normalizer = None,
                 stemmer = None, stopwords = None,
                 expander = None):
        """ The fulltext indexer
        |basename| is the basename to create the text databases
        |cachesize| Size of the memory cache (greater is better)
        |mode| Mode of created databases (permission access)
        |wordreg| Is the regular expression to read words, If is None
        letter and digits and the word length must be between 1 and 16
        characters
        |normalizer| Is the word and character normalizer, is applied before
        the stemmer. The normalizer is a function that receives every word
        (str) and returns a normalized word (str). If None the normalize
        function in the module is used.
        |stemmer| Is for every word to create stemmed thesaurous.
        The default is None
        |stopwords| A dict with the stopwords (will not be indexed). It must
        be stemmed because they are compared with the stemmed words
        |expander| Is a function to expand a word into several more words
        is called after the stemmer, normalizer and stop word filter. If None is
        given no expansion is performed"""
        # db = bsddb._db
        mode = 0644
        self.stemmer = stemmer
        self.expander = expander
        self.stopwords = stopwords
        self.__closed = False
        if wordreg is None:
            wordreg = sre.compile(r"[\d\wñÑáéíóúáäëïöüàèìòùÁÉÍÓÚÄËÏÖÜÀÈÌÒÙ]{1,16}",sre.L|sre.I)
        self.wordreg = wordreg
        if normalizer is None: normalizer = normalize
        self.normalizer = normalizer

    def __del__(self):
        if not self.__closed: self.close()

    def close(self):
        if self.__closed: return
        self.__closed = True

    def sync(self):
        pass
    
    def vectorizechunkedandappend(self, url, kwargs_vector, vecdoc=None):
        """ Do a vectorizechunked and append the vector. return None """
        vecdoc = self.vectorizechunked(kwargs_vector, vecdoc)
        self.appendvector(url, vecdoc)
        
    def vectorizechunked(self, kwargs_vector, vecdoc=None):
        """ A chunked version of vectorize.
        |kwargs_vector| is a sequence of dictionaries with
        the arguments to be passed to vectorize
        |datafile| must be named as 'data' and vecdoc is internal to
        vectorizechunked
        """
        for kwargs in kwargs_vector:
            #print "x" * 30, ">",kwargs.keys()
            data = kwargs.pop('data')
            vecdoc = self.vectorize(data,vecdoc,**kwargs)
        return vecdoc

    def vectorize(self, datafile, vecdoc = None,
                  wordreg = None, normalizer = None,
                  stemmer = None, stopwords = None,
                  expander = None):
        """ Vectorize a document
        |datafile| Is the documents data
        |vecdoc| The VectorDocument. if |vecdoc| is None a new one is
        created, if |vecdoc| is not None then append to vecdoc
        the |datafile| vector. Useful to chunked vectorizations.
        |wordreg| A regular expression to read words, if None the
        indexer wordreg will be used
        |normalizer| A functions that normalize every word, if None
        the indexer wordreg will be used
        |stemmer| Is for every word to create stemmed thesaurous.
        If None is given, the indexer stemmer is used
        |stopwords| A dict with the stopwords (words that must not be
        indexed). If None is given, the indexer stopwords are used.
        Note1: The stopwords must be already stemmed
        Note2: If you need to disable any inheretable argument as
        wordreg,normalizer,stemmer, or stopwords just set to False,
        do not use None.
        |expander| Is a function to expand a word into several more words
        is called after the stemmer, normalizer and stop word filter"""
        if stemmer is None: stemmer = self.stemmer
        if expander is None: expander = self.expander
        if stopwords is None: stopwords = self.stopwords
        if vecdoc is None:
            maxfreq = 0
            vecdoc = VectorDocument()
        else:
            maxfreq = vecdoc.maxfreq
        dicdoc = vecdoc.vector
        if wordreg is None:
            wordreg = self.wordreg
        else:
            if type(wordreg) is str: wordreg = sre.compile(wordreg)
        wordreg.findall("Sabor a mí, está mal")
        if normalizer is None: normalizer = self.normalizer
        for word in wordreg.finditer(datafile):
            word = word.group()
            if normalizer:
                word = normalizer(word)
            if stemmer: word = stemmer(word)
            if stopwords and stopwords.has_key(word): continue
            if not expander:
                freq = dicdoc.get(word, 0) + 1
                maxfreq = max(freq, maxfreq)
                dicdoc[word] = freq
            else:
                for w in expander(word):
                    freq = dicdoc.get(w, 0) + 1
                    maxfreq = max(freq, maxfreq)
                    dicdoc[w] = freq
        vecdoc.maxfreq = maxfreq
        return vecdoc

