#!/usr/bin/python
#coding: utf-8;
############################################################################
#                                                                          #
#    This file 'router.py'                                                 #
#    is part of 'SPyDI: Simple Python Distributed Indexing'                #
#    Copyright (C) 2004-2005 by Eric Sadit Tellez Avila                    #
#    sadit@lsc.fie.umich.mx or donsadit@gmail.com                          #
#                                                                          #
#    This program is free software; you can redistribute it and#or modify  #
#    it under the terms of the GNU General Public License as published by  #
#    the Free Software Foundation; either version 2 of the License, or     #
#    (at your option) any later version.                                   #
#                                                                          #
#    This program is distributed in the hope that it will be useful,       #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#    GNU General Public License for more details.                          #
#                                                                          #
#    You should have received a copy of the GNU General Public License     #
#    along with this program; if not, write to the                         #
#    Free Software Foundation, Inc.,                                       #
#    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             #
############################################################################

from threadutils import ThreadPool
import random
import SPyRO as spyro
import warn
import sys
import time

__doc__ = """
Router module

The router module performs routing over a visible objects.

"""
class InputObject:
    """ An input object representation, it is used to cache route
    information """
    def __init__(self, address, time):
        """
        |address| Peer address
        |time| time to access to this object
        """
        self.address = address
        self.time = time

class ObjectStats:
    """ Object statistics, used in the router to store statistics of
    calls and routes in objects
    """
    def __init__(self):
        self.avgtime = 10 # Average time
        self.numreq = 1 # Number of requests
        self.failures = 0 # Number of failures
        self.lasttime = 0 # Time of the last request

class Local:
    """ A local router, it is just a wrapper of an object """
    def __init__(self, dist, obj = None):
        self.obj = obj

    def register(self, obj, keys):
        """ Register an object with keys """
        self.obj = obj

    def route(self, routinginfo, key, methodname, *args, **kwargs):
        """ Execute the method |methodname| with args and keyword
        arguments |kwargs| """
        if self.obj is None:
            raise Exception, "route: local router has not obj"
        f = getattr(self.obj, methodname)
        return [ f(*args, **kwargs) ]

    def routeone(self, routinginfo, key, methodname, *args, **kwargs):
        """ Execute the method |methodname| with args and keyword
        arguments |kwargs| """
        if self.obj is None:
            raise Exception, "routeone: local router has not obj"
        f = getattr(self.obj, methodname)
        return f(*args, **kwargs)

class Routing:
    """ Router class """
    TTL = 0
    QID = 1

    __spyro__ = ( 'route', 'routeone')

    def __init__(self, dist, ttl = 12):
        """ Init the routing module """
        self.dist = dist
        self.ttl = ttl
        self.tablein = {}
        self.tablestats = {}
        self.tablereq = {}
        self.register = dist.register
        self.registerbyaddress = dist.registerbyaddress

    def routeone(self, *args, **kwargs):
        """ Route the call to a one object
        def route([peerconn,spyroreq,]
                  routinginfo, keys, methodname, *args, **kwargs)
        |peerconnectionobject,spyrorequestobject|
        are SPyRO request objects, they must be present if it is called
        directly by SPyRO, it must be omited otherwise
        |keys| arguments to the router, normally keys or urls (can be None)
        |methodname| the name of the method
        
        This is an "intelligent" method, it guess if is called
        by spyro or called by an user.
        """
        x = self.prepare(True, args, kwargs)
        return x[0]
    
    def route(self, *args, **kwargs):
        """ Route the call method to a list of objects
        def route([peerconn,spyroreq,]
                  routinginfo, keys, methodname, *args, **kwargs)
        |peerconnectionobject,spyrorequestobject|
        are SPyRO request objects, they must be present if it is called
        directly by SPyRO, it must be omited otherwise
        |keys| arguments to the router, normally keys or urls (can be None)
        |methodname| the name of the method
        
        This is an "intelligent" method, it guess if is called
        by spyro or called by an user. 
        """
        x = self.prepare(False, args, kwargs)
        return x
        
    def prepare(self, onlyone, args, kwargs):
        if isinstance(args[0], spyro.PeerConnection):
            peerconn = args[0]
            spyroreq = args[1]
            args = args[2:]
        else:
            peerconn = None
            spyroreq = None

        routinginfo, keys, methodname = args[0], args[1], args[2]
        return self.execute(onlyone, peerconn, spyroreq, routinginfo, keys,
                      methodname, args[3:], kwargs)

    def execute(self, onlyone, peerconn, spyroreq, routinginfo, keys,
                methodname, args, kwargs):
        """ This is the real execution method """
        inittime = time.time()
        # States:
        # 
        # qid = routinginfo[Routing.QID]
        # 
        # peerconn | tablein[qid] | State
        # ---------+--------------+---------------------
        #     None |         None | Init request (first)
        #     None |       Exists | Not specified
        #   Exists |         None | Remote init request
        #   Exists |       Exists | In remote request
        if peerconn is None or routinginfo is None:
            qid = "%s:%.3f"%(hash(repr(args)),inittime-946080000) # now() - 30 years
            # ttl, idquery
            ttl = self.ttl
            routinginfo = [ttl, qid]
            self.tablereq[qid] = inittime
        else:
            peer = peerconn.request.getpeername()
            qid = routinginfo[Routing.QID]
            ttl = routinginfo[Routing.TTL] - 1
            routinginfo[Routing.TTL] = ttl
            if self.tablereq.has_key(qid) or self.tablein.has_key(qid):
                return []
            self.tablein[qid] = InputObject(peer, inittime)

        res = []
        self.before(methodname, res, args, kwargs)
        for obj in self.dist.objectlist(keys):
            try:
                objstat = self.tablestats[id(obj)]
            except KeyError:
                objstat = ObjectStats()
                self.tablestats[id(obj)] = objstat
            try:
                q = obj.route(routinginfo, keys, methodname, *args, **kwargs)
                res.extend(q)
                routefails = False
            except Exception, err:
                routefails = True
                warn.warn( "routing problems:", err )
                objstat.failures = objstat.failures + 1
            objstat.lastreq = time.time()
            objstat.avgtime = (objstat.avgtime + (time.time() - inittime) / 2.0) / 2
            objstat.numreq = objstat.numreq + 1
            if routefails: continue
            if onlyone: break
            
        return self.fusion(methodname, res, args, kwargs)

    def before(self, methodname, res, args, kwargs):
        pass
        
    def fusion(self, methodname, res, args, kwargs):
        """ A function that interprets to interpret and modify
        the results. It must return a list
        |methodname| Method's name
        |res| Results to be interpreted
        |args| Arguments used in the call
        |kwargs| Keyword arguments used in the call
        """
        return res

class Hub:
    def __init__(self, router, methods = None, keys = None):
        self.router = router
        self.methods = methods
        self.__spyro__ = ()
        self.keys = keys
     
    def __getattr__(self, name):
        if self.methods is None or name in self.methods:
            return lambda *args, **kwargs: self.router.route(None, self.keys, name, *args, **kwargs)
        raise AttributeError, name

if __name__ == '__main__':
    import SPyRO as spyro
    import sys, os, time    
    from threadutils import ServerInThread
    from distributor import Random

    s0 = ServerInThread(spyro.Server, ('127.0.0.1', 2000))
    s1 = ServerInThread(spyro.Server, ('127.0.0.1', 2001))
    s2 = ServerInThread(spyro.Server, ('127.0.0.1', 2002))
    s3 = ServerInThread(spyro.Server, ('127.0.0.1', 2003))

    class Bici:
        Num = 0
        def __init__(self,ident):
            Bici.Num += 10
            self.num = Bici.Num
            self.ident = ident
        
        def run(self):
            return "%s run %03d"%(self.ident, self.num)
    routers = []
    dists = []
    index = 0
    for s in [s0,s1,s2,s3]:
        dist = Random(spread=True)
        bici0 = Bici("comm%d"%index)
        bici1 = Bici("comm%d"%index)
        bici2 = Bici("comm%d"%index)
        index = index + 1
        dist.register(Local(None, bici0), None)
        dist.register(Local(None, bici1), None)
        dist.register(Local(None, bici2), None)
        dists.append(dist)
        # routers.append(StringRouting(dist))
        routers.append(Routing(dist))
        s.server.setobj("router", routers[-1])
        s.server.setobj("dist", dist)
        s.server.setobj("bici0", bici0)
        s.server.setobj("bici1", bici1)
        s.server.setobj("bici2", bici2)
        
    comm0 = spyro.TClient(("127.0.0.1",2000),"/")
    comm1 = spyro.TClient(("127.0.0.1",2001),"/")
    comm2 = spyro.TClient(("127.0.0.1",2002),"/")
    comm3 = spyro.TClient(("127.0.0.1",2003),"/")
    
    try:
        ret = routers[0].route(None, None, "run")
        print "*************A routers[0]->run ", ret
        routers[0].register(comm1.getobj("router"), None)
        routers[1].register(comm0.getobj("router"), None)
        routers[0].register(comm2.getobj("router"), None)
        routers[2].register(comm1.getobj("router"), None)
        routers[1].register(comm3.getobj("router"), None)
        routers[1].register(comm0.getobj("router"), None)
        print "=============  registered routers[1] to routers[0]"
        ret = routers[0].route(None, None, "run")
        print "*************B routers[0]->run ", repr(ret)
        hub = Hub(routers[0])
        print "*************C ", hub.run()
    except:
        warn.warn()
    
    for d in dists:
        d.close()

    for comm in [comm0,comm1,comm2,comm3]:
        comm.close()

    for s in [s0,s1,s2,s3]:
        s.close()
